//
//
// Brunelco
//
// Kernel Mode Treiber (intel only)
//
// fuer 8-Channel PC-Timer Card von Brunelco
//
#include "ntddk.h"
#include "brun.h"

//  Debug-Ausgabe erfolgt nur im "Checked-Build"
#if DBG
#define DbgPrint(arg) DbgPrint arg
#else
#define DbgPrint(arg)
#endif

ULONG GlobalBaseAddress=0x320;
ULONG MaxLoops=500000;

//
// Brunelco_DispatchCreateClose
//
// wird bei CreateFile und CloseHandle aufgerufen und gibt
// einfach STATUS_SUCCESS zurueck
//
NTSTATUS Brunelco_DispatchCreateClose(IN PDEVICE_OBJECT devObj,IN PIRP Irp)

{
    DbgPrint(("Brunelco: Open / Close\n")); 
    Irp->IoStatus.Information = 0;
    Irp->IoStatus.Status = STATUS_SUCCESS;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return STATUS_SUCCESS;
}


//
// Brunelco_DispatchRead
//
// wird bei ReadFile vom I/O Manager aufgerufen
//
// ACHTUNG: GlobaleBaseAdress muss einen richtigen Wert haben !!!
//          Voreingestellt ist 0x320. Dieser Wert kann ber
//          DeviceIOControl mit dem Controlcode 
//          IOCTL_Brunelco_SetBaseAddress veraendert werden
//
NTSTATUS Brunelco_DispatchRead(IN PDEVICE_OBJECT devObj,IN PIRP Irp)
{
    UCHAR *mBuffer;
    PIO_STACK_LOCATION IrpStack = IoGetCurrentIrpStackLocation(Irp);

    if (IrpStack->Parameters.Read.Length>=6)
      {
       mBuffer=Irp->AssociatedIrp.SystemBuffer;

       _asm
          {     
                mov   ebx,0
                mov   edx,GlobalBaseAddress
                mov   esi,mBuffer
             loop1:    
                inc   ebx
                mov   ecx,MaxLoops         // max MaxLoops Durchlaeufe
                inc   edx
                mov   eax,ebx
                out   dx,al                // Funktion festlegen
                dec   edx
             loop0:   
                in    al,dx
                test  al,al                // Test ob
                jns   ok0                  // Bit 7 gesetzt ist
                dec   ecx  
                jnz   loop0
                mov   byte ptr [esi],128   // *mBuffer=128
                jmp   exit0                // break
             ok0:
                in    al,dx
                mov   byte ptr [esi],al    // *mBuffer=READ_PORT_UCHAR()
                
                inc   esi                  // mBuffer++
                cmp   ebx,6                // Funktionscode=6 ?
                jne   loop1                // nein - dann -> loop1
                
             exit0:   
          }
       
       Irp->IoStatus.Information=6;
       Irp->IoStatus.Status=STATUS_SUCCESS;
      } else
      {
       Irp->IoStatus.Information = 0;
       Irp->IoStatus.Status = STATUS_BUFFER_TOO_SMALL;
      }

    IoCompleteRequest(Irp,IO_NO_INCREMENT);
    return STATUS_SUCCESS;
}


//
// Brunelco_DispatchDeviceControl
//
// wird bei DeviceIOControl vom I/O Manager aufgerufen
//
// Control Codes:
//
// IOCTL_Brunelco_UniAccess      - universeller Zugriff auf alle Funktionen
//                                 ueber mParamBlock wird die Basisadresse
//                                 und der Funktionscode festgelegt
// IOCTL_Brunelco_GetLastTime    - liefert letzten gespeicherten Zeitwert
// IOCTL_Brunelco_SetBaseAddress - veraendert die Basisadresse
// IOCTL_Brunelco_GetVersion     - liefert die Version des Treibers
//
NTSTATUS Brunelco_DispatchDeviceControl(IN PDEVICE_OBJECT devObj,IN PIRP Irp)

{

struct mParamBlockStruct
     {
      ULONG mBaseAdr;           // Basisadresse
      ULONG mFunctionCode;      // Funktionsnummer
     } *mParamBlock;

    UCHAR *mBuffer,*mVer=mVersion;
    ULONG mInputBufferLength, mOutputBufferLength, dummy, mLoopCount;
    ULONG mIOCTLCode,mFunctionCode;
    NTSTATUS sts;
    
    PIO_STACK_LOCATION IrpStack = IoGetCurrentIrpStackLocation(Irp);
    
    mIOCTLCode=IrpStack->Parameters.DeviceIoControl.IoControlCode;
    mInputBufferLength=IrpStack->Parameters.DeviceIoControl.InputBufferLength;
    mOutputBufferLength=IrpStack->Parameters.DeviceIoControl.OutputBufferLength;

    DbgPrint(("Brunelco: InputBufferLength = %x\n",mInputBufferLength));
    DbgPrint(("Brunelco: OutputBufferLength = %x\n",mOutputBufferLength));
    DbgPrint(("Brunelco: FunctionCode = %x\n",mIOCTLCode));
    
//
// IOCTL_Brunelco_UniAccess
//
    if (mIOCTLCode==IOCTL_Brunelco_UniAccess)
      {
        mParamBlock=Irp->AssociatedIrp.SystemBuffer;
        dummy=mParamBlock->mBaseAdr;

        DbgPrint(("Brunelco: UniAccess - Adresse %x Funktionnummer %x\n",
                  dummy,mParamBlock->mFunctionCode));
                  
        //
        // Funktionscode schreiben
        //
        WRITE_PORT_UCHAR(++dummy,mParamBlock->mFunctionCode);
        mBuffer=Irp->AssociatedIrp.SystemBuffer;
        dummy--;
        mLoopCount=0;
        //
        // Rueckgabe lesen
        //
        do
          { 
           *mBuffer=READ_PORT_UCHAR(dummy);
           mLoopCount++;
          } while ((*mBuffer>127)&&(mLoopCount<MaxLoops));
        //
        // bei Timeout nach MaxLoops - Rueckgabe des Wertes 128
        //  
        if (mLoopCount<MaxLoops) *mBuffer=READ_PORT_UCHAR(dummy);
           else *mBuffer=128;
        //
        // Laenge der Rueckgabedaten ist immer 1
        //   
        Irp->IoStatus.Information = 1;
        DbgPrint(("Brunelco: Adresse %x - Gelesen %x\n",dummy,*mBuffer));
      } else
//
// IOCTL_Brunelco_GetLastTime
//
     if (mIOCTLCode==IOCTL_Brunelco_GetLastTime)
      {
        //
        // ist OutputBuffer gross genug ?
        //
        if (mOutputBufferLength<6)
          {
            //
            // nein, dann Anfrage abschliessen und als
            // Status STATUS_BUFFER_TOO_SMALL zurueckgeben
            //
            Irp->IoStatus.Information = 0;
            Irp->IoStatus.Status = STATUS_BUFFER_TOO_SMALL;
            IoCompleteRequest(Irp, IO_NO_INCREMENT);
            return STATUS_BUFFER_TOO_SMALL;
          }

        mParamBlock=Irp->AssociatedIrp.SystemBuffer;
        dummy=mParamBlock->mBaseAdr;
        DbgPrint(("Brunelco: GetLastTime - Adresse %x \n",dummy));
        mBuffer=Irp->AssociatedIrp.SystemBuffer;

        //
        // Funktionen 1 - 6 ausfuehren
        //
        for (mFunctionCode=1;mFunctionCode<7;mFunctionCode++)
          {
           mLoopCount=0;
           WRITE_PORT_UCHAR(++dummy,mFunctionCode);
           dummy--;
           do
             { 
              *mBuffer=READ_PORT_UCHAR(dummy);
              mLoopCount++;
             } while ((*mBuffer>127)&&(mLoopCount<MaxLoops));
           if (mLoopCount<MaxLoops) *mBuffer=READ_PORT_UCHAR(dummy);
              else *mBuffer=128;
           mBuffer++;
          }
        Irp->IoStatus.Information = 6;
      } else
//
// IOCTL_Brunelco_SetBaseAddress
//
    if (mIOCTLCode==IOCTL_Brunelco_SetBaseAddress)
      {
        mParamBlock=Irp->AssociatedIrp.SystemBuffer;
        GlobalBaseAddress=mParamBlock->mBaseAdr;
        Irp->IoStatus.Information = 0;
      } else
//
// IOCTL_Brunelco_GetVersion
//
    if (mIOCTLCode==IOCTL_Brunelco_GetVersion)
      {
        if (mOutputBufferLength<6)
          {
            Irp->IoStatus.Information = 0;
            Irp->IoStatus.Status = STATUS_BUFFER_TOO_SMALL;
            IoCompleteRequest(Irp, IO_NO_INCREMENT);
            return STATUS_BUFFER_TOO_SMALL;
          }
        mBuffer=Irp->AssociatedIrp.SystemBuffer;
        RtlCopyBytes(mBuffer,mVer,6);
        Irp->IoStatus.Information = 6;
       }  
    
    Irp->IoStatus.Status = STATUS_SUCCESS;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return STATUS_SUCCESS;
}

//
// Brunelco_Unload
//
// ermoeglicht das Entladen des Treibers
// 
// der symbolische Link wird entfernt und das
// Geraeteobjekt geloescht
//

VOID Brunelco_Unload( IN PDRIVER_OBJECT DrvObj)
{
     WCHAR LinkName[]=L"\\DosDevices\\Brunelco";
     UNICODE_STRING uniLinkName;
     NTSTATUS sts;     

     DbgPrint(("Brunelco: Eintritt in Unload Routine !\n"));
     RtlInitUnicodeString(&uniLinkName, LinkName);
     sts=IoDeleteSymbolicLink(&uniLinkName);
     if (!NT_SUCCESS(sts))
      {
        DbgPrint(("Brunelco: Kann Link nicht entfernen !\n"));
        return;
      } else DbgPrint(("Brunelco: Link entfernt !\n"));
     IoDeleteDevice(DrvObj->DeviceObject);
     DbgPrint(("Brunelco: Unload Ende !\n"));
}


//
// DriverEntry
//
// wird beim Laden des Treibers vom I/O Manager aufgerufen
//
// die Dispatch Routinen werden im Treiberobjekt eingetragen
// die Unload Routine wird registriert
// das Geraeteobjekt und die symbolische Verknuepfung werden erzeugt
// 

NTSTATUS DriverEntry( IN PDRIVER_OBJECT drvObj,
                      IN PUNICODE_STRING RegistryPath)

{
     PDEVICE_OBJECT devObj;
     NTSTATUS sts;
     WCHAR Name[]=L"\\Device\\Brunelco";
     WCHAR LinkName[]=L"\\DosDevices\\Brunelco";
     UNICODE_STRING uniName, uniLinkName;

     RtlInitUnicodeString(&uniName, Name);
     RtlInitUnicodeString(&uniLinkName, LinkName);
     
     drvObj->MajorFunction[IRP_MJ_CREATE]=Brunelco_DispatchCreateClose;
     drvObj->MajorFunction[IRP_MJ_CLOSE]=Brunelco_DispatchCreateClose;
     drvObj->MajorFunction[IRP_MJ_DEVICE_CONTROL]=Brunelco_DispatchDeviceControl;
     drvObj->MajorFunction[IRP_MJ_READ]=Brunelco_DispatchRead;
     drvObj->DriverUnload=Brunelco_Unload;
     
     //
     // Geraeteobjekt erzeugen
     //
     sts=IoCreateDevice(drvObj,0,&uniName,FILE_DEVICE_UNKNOWN,0,TRUE,&devObj);
     if (!NT_SUCCESS(sts))
      {
        DbgPrint(("Brunelco: Kann DeviceObject nicht erstellen !\n"));
        return sts;
      } 

     //
     // symbolischen Link erzeugen
     //
     sts=IoCreateSymbolicLink(&uniLinkName, &uniName);
     if (!NT_SUCCESS(sts))
      {
        DbgPrint(("Brunelco: Kann Link nicht erstellen !\n"));
        return sts;
      } else
      {
        DbgPrint(("Brunelco: Name: %s - Link: %s\n", Name, LinkName));
      } 

    //
    // der Treiber verwendet Buffered I/O
    //
    devObj->Flags |= DO_BUFFERED_IO;
    
    DbgPrint(("Brunelco: Driver Entry erfolgreich abgearbeitet !\n"));
    DbgPrint(("Brunelco: FunctionCodes: "));

    return STATUS_SUCCESS;
}     

